/****************************************************************************
** Demonstration of moving small bitmaps (not _real_ sprites, but the next **
** best thing) around the screen                                           **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include "malloc.h"
#include "dos.h"
#include "mem.h"
#ifdef __DJGPP
  #include "sys/nearptr.h"
#endif

struct SpriteRec {
  //These indicate the dimensions of the sprite
  unsigned int Width, Height, Total;
  //These indicate the location of the sprite in system memory
  #ifdef __DJGPP
    unsigned char *SpriteData;
  #else
    unsigned char far *SpriteData;
  #endif

  //NOTE: The following values are optional

  //These indicate the location of the sprite on screen
  unsigned int X, Y;
  //These indicate the direction of the sprite movement
  char Xdir, Ydir;
};

  struct SpriteRec   Sprite1;
  struct SpriteRec   Sprite2;
  unsigned int       tX, tY;

  #ifdef __DJGPP
    unsigned char Buffer [64000];
    unsigned char *Screen;
  #else
    unsigned char far *Buffer;
    unsigned char far *Screen;
  #endif

//Switches to VGA 320x200 256 colour
void VGAScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x13;
  int86(0x10, &r, &r);
}

//Switches to text 80x25
void TextScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x3;
  int86(0x10, &r, &r);
}

//Sets a pixel in the buffer to a specific colour
void PutPixel (unsigned int x, unsigned int y, unsigned char c)
{
  if ((x < 320) && (y < 200)) {
    Buffer [(y << 8) + (y << 6) + x] = c;
  }
}

//Reads the colour value of a pixel from the buffer
unsigned char ReadPixel (unsigned int x, unsigned int y)
{
  if ((x < 320) && (y < 200)) {
    return (Buffer [(y << 8) + (y << 6) + x]);
  } else {
    return (0);
  }
}

//Clears the buffer
void ClearBuffer ()
{
  #ifdef __DJGPP
    memset (Buffer, 0, 64000);
  #else
    _fmemset (Buffer, 0, 64000);
  #endif
}

//Displays the buffer
void DisplayBuffer ()
{
  #ifdef __DJGPP
    memcpy (Screen, Buffer, 64000);
  #else
    _fmemcpy (Screen, Buffer, 64000);
  #endif
}

unsigned char MakeSprite (struct SpriteRec *Sprite,
                          unsigned int x1, unsigned int y1,
                          unsigned int x2, unsigned int y2)

{
  unsigned int      Size, x, y;
  unsigned int      CurrentPixel;
  #ifdef __DJGPP
    unsigned char *SpriteMem;
  #else
    unsigned char far *SpriteMem;
  #endif

  //Calculate amount of memory needed for sprite
  Size = (y2 - y1 + 1) * (x2 - x1 + 1) - 1;
  //Check if enough memory is available
  #ifdef __DJGPP
    if ((SpriteMem = (char *) malloc(Size))==NULL) {
  #else
    if ((SpriteMem = (char far *) farmalloc(Size))==NULL) {
  #endif
    //If not, return 0
    return (0);
  }
  Sprite->SpriteData = SpriteMem;

  //Set dimensions
  Sprite->Width   = x2 - x1;
  Sprite->Height  = y2 - y1;
  Sprite->Total   = Size + 1;
  //Increase width/height for correct loop
  x2++; y2++;
  //Read sprite from the screen buffer
  CurrentPixel = 0;
  for (y = y1; y < y2; y++) {
    for (x = x1; x < x2; x++) {
      Sprite->SpriteData [CurrentPixel++] =
        ReadPixel (x, y);
    }
  }
  //Return true
  return (1);
}

void PutSprite (struct SpriteRec Sprite, unsigned int x, unsigned int y)
{
  unsigned int  XOfs, YOfs, CurrentPixel;
  unsigned char DataByte;

  //Set location to top left corner of the sprite
  XOfs = 0; YOfs = 0;
  //Proces the entire sprite
  for (CurrentPixel = 0; CurrentPixel < Sprite.Total; CurrentPixel++) {
    //Read pixel from sprite
    DataByte = Sprite.SpriteData [CurrentPixel];
    //Check to see whether the pixel is transparent
    if (DataByte > 0) PutPixel (XOfs + x, YOfs + y, DataByte);
    //Set pointers to next position
    XOfs++;
    //Switch to next line if the end of the current line has been reached
    if (XOfs > Sprite.Width) {
      XOfs = 0;
      YOfs++;
    }
  }
}

void main () {
  //Switch to graphics mode
  VGAScreen ();

  #ifdef __DJGPP
    __djgpp_nearptr_enable ();
    Screen = (char *) (__djgpp_conventional_base + 0xA0000);
  #else
    Screen = (unsigned char *)MK_FP (0xA000, 0000);
    //Assign memory to buffer
    if ((Buffer = (unsigned char far *) farmalloc(64000))==NULL) {
      //If there wasn't enough memory, quit
      TextScreen ();
      printf ("Not enough memory for frame buffer!\n");
      return;
    }
  #endif

  //Clear the buffer
  ClearBuffer ();
  //Set up a sprite consisting of a blue block
  for (tX = 0; tX < 16; tX++) {
    for (tY = 0; tY < 16; tY++) {
      PutPixel (tX, tY, 1);
    }
  }

  //Get the sprite
  if (!MakeSprite (&Sprite1, 0, 0, 15, 15)) {
    //If there wasn't enough memory, quit
    TextScreen ();
    printf ("Not enough memory to create sprite!\n");
    return;
  }

  //Set up a sprite consisting of a red rectangle
  for (tX = 0; tX < 16; tX++) {
    PutPixel (16 + tX,  0, 4);
    PutPixel (16 + tX, 15, 4);
    PutPixel (16 +  0, tX, 4);
    PutPixel (16 + 15, tX, 4);
  }
  //Get the sprite
  if (!MakeSprite (&Sprite2, 16, 0, 16 + 15, 15)) {
    //If there wasn't enough memory, quit
    TextScreen ();
    printf ("Not enough memory to create sprite!\n");
    return;
  }

  //Set up sprite1 coordinates at top left of the screen
  Sprite1.X = 0;
  Sprite1.Y = 0;
  //Set direction down right
  Sprite1.Xdir = 1;
  Sprite1.Ydir = 1;

  //Set up sprite2 coordinates at bottom left of the screen
  Sprite2.X = 0;
  Sprite2.Y = 184;
  //Set direction up right
  Sprite2.Xdir = 1;
  Sprite2.Ydir = -1;

  do {
    //Clear the buffer - if you leave this out, the sprites will leave trails
    ClearBuffer ();

    //Put the first sprite in the buffer
    PutSprite (Sprite1, Sprite1.X, Sprite1.Y);
    //Put the second sprite in the buffer
    PutSprite (Sprite2, Sprite2.X, Sprite2.Y);

    //Display the buffer - if you leave this out, you won't see a thing
    DisplayBuffer ();

    //Update first sprite location
    Sprite1.X += Sprite1.Xdir;
    Sprite1.Y += Sprite1.Ydir;
    //Bounce off the sides
    if (Sprite1.X < 1) Sprite1.Xdir = 1;
    if (Sprite1.X > 304) Sprite1.Xdir = -1;
    if (Sprite1.Y < 1) Sprite1.Ydir = 1;
    if (Sprite1.Y > 184) Sprite1.Ydir = -1;

    //Update second sprite location
    Sprite2.X += Sprite2.Xdir;
    Sprite2.Y += Sprite2.Ydir;
    //Bounce off the sides
    if (Sprite2.X < 1) Sprite2.Xdir = 1;
    if (Sprite2.X > 304) Sprite2.Xdir = -1;
    if (Sprite2.Y < 1) Sprite2.Ydir = 1;
    if (Sprite2.Y > 184) Sprite2.Ydir = -1;
  } while (!kbhit());
  getch ();
  TextScreen ();
}